@@ -12,7 +12,7 @@ from __future__ import absolute_import
 #
 # Imports
 #
-
+import copy
 import errno
 import itertools
 import os
@@ -44,10 +44,10 @@ from .exceptions import (
     WorkerLostError,
 )
 from .five import Empty, Queue, range, values, reraise, monotonic
-from .util import Finalize, debug
+from .util import Finalize, debug, warning
 
 MAXMEM_USED_FMT = """\
-child process exiting after exceeding memory limit ({0}KiB / {0}KiB)
+child process exiting after exceeding memory limit ({0}KiB / {1}KiB)
 """
 
 PY3 = sys.version_info[0] == 3
@@ -110,6 +110,8 @@ SIG_SOFT_TIMEOUT = getattr(signal, "SIGUSR1", None)
 
 LOST_WORKER_TIMEOUT = 10.0
 EX_OK = getattr(os, "EX_OK", 0)
+GUARANTEE_MESSAGE_CONSUMPTION_RETRY_LIMIT = 300
+GUARANTEE_MESSAGE_CONSUMPTION_RETRY_INTERVAL = 0.1
 
 job_counter = itertools.count()
 
@@ -135,8 +137,7 @@ def starmapstar(args):
 
 
 def error(msg, *args, **kwargs):
-    if util._logger:
-        util._logger.error(msg, *args, **kwargs)
+    util.get_logger().error(msg, *args, **kwargs)
 
 
 def stop_if_not_current(thread, timeout=None):
@@ -237,7 +238,7 @@ class Worker(object):
     def __init__(self, inq, outq, synq=None, initializer=None, initargs=(),
                  maxtasks=None, sentinel=None, on_exit=None,
                  sigprotection=True, wrap_exception=True,
-                 max_memory_per_child=None):
+                 max_memory_per_child=None, on_ready_counter=None):
         assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)
         self.initializer = initializer
         self.initargs = initargs
@@ -248,6 +249,7 @@ class Worker(object):
         self.sigprotection = sigprotection
         self.inq, self.outq, self.synq = inq, outq, synq
         self.wrap_exception = wrap_exception  # XXX cannot disable yet
+        self.on_ready_counter = on_ready_counter
         self.contribute_to_object(self)
 
     def contribute_to_object(self, obj):
@@ -269,7 +271,7 @@ class Worker(object):
         return self.__class__, (
             self.inq, self.outq, self.synq, self.initializer,
             self.initargs, self.maxtasks, self._shutdown, self.on_exit,
-            self.sigprotection, self.wrap_exception,
+            self.sigprotection, self.wrap_exception, self.max_memory_per_child,
         )
 
     def __call__(self):
@@ -278,7 +280,7 @@ class Worker(object):
 
         def exit(status=None):
             _exitcode[0] = status
-            return _exit()
+            return _exit(status)
         sys.exit = exit
 
         pid = os.getpid()
@@ -344,49 +346,70 @@ class Worker(object):
                 i += 1
 
         completed = 0
-        while maxtasks is None or (maxtasks and completed < maxtasks):
-            req = wait_for_job()
-            if req:
-                type_, args_ = req
-                assert type_ == TASK
-                job, i, fun, args, kwargs = args_
-                put((ACK, (job, i, now(), pid, synqW_fd)))
-                if _wait_for_syn:
-                    confirm = wait_for_syn(job)
-                    if not confirm:
-                        continue  # received NACK
-                try:
-                    result = (True, prepare_result(fun(*args, **kwargs)))
-                except Exception:
-                    result = (False, ExceptionInfo())
-                try:
-                    put((READY, (job, i, result, inqW_fd)))
-                except Exception as exc:
-                    _, _, tb = sys.exc_info()
+        try:
+            while maxtasks is None or (maxtasks and completed < maxtasks):
+                req = wait_for_job()
+                if req:
+                    type_, args_ = req
+                    assert type_ == TASK
+                    job, i, fun, args, kwargs = args_
+                    put((ACK, (job, i, now(), pid, synqW_fd)))
+                    if _wait_for_syn:
+                        confirm = wait_for_syn(job)
+                        if not confirm:
+                            continue  # received NACK
                     try:
-                        wrapped = MaybeEncodingError(exc, result[1])
-                        einfo = ExceptionInfo((
-                            MaybeEncodingError, wrapped, tb,
-                        ))
-                        put((READY, (job, i, (False, einfo), inqW_fd)))
-                    finally:
-                        del(tb)
-                completed += 1
-                if max_memory_per_child > 0:
-                    used_kb = mem_rss()
-                    if used_kb <= 0:
-                        error('worker unable to determine memory usage')
-                    if used_kb > 0 and used_kb > max_memory_per_child:
-                        error(MAXMEM_USED_FMT.format(
-                            used_kb, max_memory_per_child))
-                        return EX_RECYCLE
-                    else:
-                        error('worker unable to determine worker memory usage')
+                        result = (True, prepare_result(fun(*args, **kwargs)))
+                    except Exception:
+                        result = (False, ExceptionInfo())
+                    try:
+                        put((READY, (job, i, result, inqW_fd)))
+                    except Exception as exc:
+                        _, _, tb = sys.exc_info()
+                        try:
+                            wrapped = MaybeEncodingError(exc, result[1])
+                            einfo = ExceptionInfo((
+                                MaybeEncodingError, wrapped, tb,
+                            ))
+                            put((READY, (job, i, (False, einfo), inqW_fd)))
+                        finally:
+                            del(tb)
+                    completed += 1
+                    if max_memory_per_child > 0:
+                        used_kb = mem_rss()
+                        if used_kb <= 0:
+                            error('worker unable to determine memory usage')
+                        if used_kb > 0 and used_kb > max_memory_per_child:
+                            warning(MAXMEM_USED_FMT.format(
+                                used_kb, max_memory_per_child))
+                            return EX_RECYCLE
 
-        debug('worker exiting after %d tasks', completed)
-        if maxtasks:
-            return EX_RECYCLE if completed == maxtasks else EX_FAILURE
-        return EX_OK
+            debug('worker exiting after %d tasks', completed)
+            if maxtasks:
+                return EX_RECYCLE if completed == maxtasks else EX_FAILURE
+            return EX_OK
+        finally:
+            # Before exiting the worker, we want to ensure that that all
+            # messages produced by the worker have been consumed by the main
+            # process. This prevents the worker being terminated prematurely
+            # and messages being lost.
+            self._ensure_messages_consumed(completed=completed)
+
+    def _ensure_messages_consumed(self, completed):
+        """ Returns true if all messages sent out have been received and
+        consumed within a reasonable amount of time """
+
+        if not self.on_ready_counter:
+            return False
+
+        for retry in range(GUARANTEE_MESSAGE_CONSUMPTION_RETRY_LIMIT):
+            if self.on_ready_counter.value >= completed:
+                debug('ensured messages consumed after %d retries', retry)
+                return True
+            time.sleep(GUARANTEE_MESSAGE_CONSUMPTION_RETRY_INTERVAL)
+        warning('could not ensure all messages were consumed prior to '
+                'exiting')
+        return False
 
     def after_fork(self):
         if hasattr(self.inq, '_writer'):
@@ -676,7 +699,11 @@ class TimeoutHandler(PoolThread):
     def _trywaitkill(self, worker):
         debug('timeout: sending TERM to %s', worker._name)
         try:
-            worker.terminate()
+            if os.getpgid(worker.pid) == worker.pid:
+                debug("worker %s is a group leader. It is safe to kill (SIGTERM) the whole group", worker.pid)
+                os.killpg(os.getpgid(worker.pid), signal.SIGTERM)
+            else:
+                worker.terminate()
         except OSError:
             pass
         else:
@@ -684,12 +711,15 @@ class TimeoutHandler(PoolThread):
                 return
         debug('timeout: TERM timed-out, now sending KILL to %s', worker._name)
         try:
-            _kill(worker.pid, SIGKILL)
+            if os.getpgid(worker.pid) == worker.pid:
+                debug("worker %s is a group leader. It is safe to kill (SIGKILL) the whole group", worker.pid)
+                os.killpg(os.getpgid(worker.pid), signal.SIGKILL)
+            else:
+                _kill(worker.pid, SIGKILL)
         except OSError:
             pass
 
     def handle_timeouts(self):
-        cache = self.cache
         t_hard, t_soft = self.t_hard, self.t_soft
         dirty = set()
         on_soft_timeout = self.on_soft_timeout
@@ -703,12 +733,16 @@ class TimeoutHandler(PoolThread):
 
         # Inner-loop
         while self._state == RUN:
+            # Perform a shallow copy before iteration because keys can change.
+            # A deep copy fails (on shutdown) due to thread.lock objects.
+            # https://github.com/celery/billiard/issues/260
+            cache = copy.copy(self.cache)
 
             # Remove dirty items not in cache anymore
             if dirty:
                 dirty = set(k for k in dirty if k in cache)
 
-            for i, job in list(cache.items()):
+            for i, job in cache.items():
                 ack_time = job._time_accepted
                 soft_timeout = job._soft_timeout
                 if soft_timeout is None:
@@ -745,7 +779,7 @@ class ResultHandler(PoolThread):
 
     def __init__(self, outqueue, get, cache, poll,
                  join_exited_workers, putlock, restart_state,
-                 check_timeouts, on_job_ready):
+                 check_timeouts, on_job_ready, on_ready_counters=None):
         self.outqueue = outqueue
         self.get = get
         self.cache = cache
@@ -757,6 +791,7 @@ class ResultHandler(PoolThread):
         self._shutdown_complete = False
         self.check_timeouts = check_timeouts
         self.on_job_ready = on_job_ready
+        self.on_ready_counters = on_ready_counters
         self._make_methods()
         super(ResultHandler, self).__init__()
 
@@ -785,6 +820,14 @@ class ResultHandler(PoolThread):
                 item = cache[job]
             except KeyError:
                 return
+
+            if self.on_ready_counters:
+                worker_pid = next(iter(item.worker_pids()), None)
+                if worker_pid and worker_pid in self.on_ready_counters:
+                    on_ready_counter = self.on_ready_counters[worker_pid]
+                    with on_ready_counter.get_lock():
+                        on_ready_counter.value += 1
+
             if not item.ready():
                 if putlock is not None:
                     putlock.release()
@@ -992,16 +1035,11 @@ class Pool(object):
         if on_process_exit is not None and not callable(on_process_exit):
             raise TypeError('on_process_exit must be callable')
 
-        class Process(self._ctx.Process):
-            _controlled_termination = False
-
-            def terminate_controlled(self):
-                self._controlled_termination = True
-                self.terminate()
-        self._Process = Process
+        self._Process = self._ctx.Process
 
         self._pool = []
         self._poolctrl = {}
+        self._on_ready_counters = {}
         self.putlocks = putlocks
         self._putlock = semaphore or LaxBoundedSemaphore(self._processes)
         for i in range(self._processes):
@@ -1067,7 +1105,8 @@ class Pool(object):
             self._outqueue, self._quick_get, self._cache,
             self._poll_result, self._join_exited_workers,
             self._putlock, self.restart_state, self.check_timeouts,
-            self.on_job_ready, **extra_kwargs
+            self.on_job_ready, on_ready_counters=self._on_ready_counters,
+            **extra_kwargs
         )
 
     def on_job_ready(self, job, i, obj, inqW_fd):
@@ -1100,6 +1139,7 @@ class Pool(object):
     def _create_worker_process(self, i):
         sentinel = self._ctx.Event() if self.allow_restart else None
         inq, outq, synq = self.get_process_queues()
+        on_ready_counter = self._ctx.Value('i')
         w = self.WorkerProcess(self.Worker(
             inq, outq, synq, self._initializer, self._initargs,
             self._maxtasksperchild, sentinel, self._on_process_exit,
@@ -1108,6 +1148,7 @@ class Pool(object):
             sigprotection=self.threads,
             wrap_exception=self._wrap_exception,
             max_memory_per_child=self._max_memory_per_child,
+            on_ready_counter=on_ready_counter,
         ))
         self._pool.append(w)
         self._process_register_queues(w, (inq, outq, synq))
@@ -1116,6 +1157,7 @@ class Pool(object):
         w.index = i
         w.start()
         self._poolctrl[w.pid] = sentinel
+        self._on_ready_counters[w.pid] = on_ready_counter
         if self.on_process_up:
             self.on_process_up(w)
         return w
@@ -1166,6 +1208,7 @@ class Pool(object):
                 self.process_flush_queues(worker)
                 del self._pool[i]
                 del self._poolctrl[worker.pid]
+                del self._on_ready_counters[worker.pid]
         if cleaned:
             all_pids = [w.pid for w in self._pool]
             for job in list(self._cache.values()):
